import * as THREE from "three";
import { OrbitControls } from "three/examples/jsm/controls/OrbitControls";
import fragShaderExample from "./shaders/grad.frag";
import vertShaderExample from "./shaders/grad.vert";
import Stats from "three/examples/jsm/libs/stats.module";

class TestClass {
    settings = {
        fov: 75,
    };
    renderer = new THREE.WebGLRenderer({
        antialias: true,
    });
    scene = new THREE.Scene();
    camera: THREE.PerspectiveCamera;
    controls: OrbitControls;
    stats = Stats();
    sceneWidth: number;
    sceneHeight: number;
    colorBoard: THREE.Mesh<THREE.BufferGeometry, THREE.ShaderMaterial> = new THREE.Mesh();

    constructor() {
        const container = document.querySelector(".canvas-wrap");
        this.sceneWidth = window.innerWidth;
        this.sceneHeight = window.innerHeight;

        // レンダラー
        this.renderer.setSize(this.sceneWidth, this.sceneHeight);
        this.renderer.setPixelRatio(window.devicePixelRatio);
        this.renderer.physicallyCorrectLights = true;
        this.renderer.shadowMap.enabled = true;
        this.renderer.shadowMap.type = THREE.PCFSoftShadowMap;
        container?.appendChild(this.renderer.domElement);

        // シーン
        this.scene.background = new THREE.Color(0x606060);

        // カメラ
        this.camera = new THREE.PerspectiveCamera(
            this.settings.fov, this.sceneWidth / this.sceneHeight, 1, 1000
        );
        this.camera.position.set(30, 20, 30);

        // コントロール
        this.controls = new OrbitControls(this.camera, this.renderer.domElement);
        this.controls.maxPolarAngle = Math.PI / 2;

        // ウィンドウリサイズ対応
        window.addEventListener("resize", () => {
            this.sceneWidth = window.innerWidth;
            this.sceneHeight = window.innerHeight;
            // レンダラーの設定変更
            this.renderer.setSize(this.sceneWidth, this.sceneHeight);
            // カメラの設定変更
            this.camera.aspect = this.sceneWidth / this.sceneHeight;
            this.camera.updateProjectionMatrix();
        }, false);

        // FPS表示
        this.stats.showPanel(0);
        container?.appendChild(this.stats.dom);

        // シーンの構成
        this.buildScene();
    }

    buildScene() {
        // 平面
        const planeGeometry = new THREE.PlaneGeometry(500, 500);
        const planeMaterial = new THREE.MeshPhongMaterial();
        const plane = new THREE.Mesh(planeGeometry, planeMaterial);
        plane.rotateX(-Math.PI / 2);
        plane.receiveShadow = true;
        this.scene.add(plane);

        // 光源
        const ambientLight = new THREE.AmbientLight(0x404040);
        this.scene.add(ambientLight);
        this.addbulbLight(-15, 15, 10);
        this.addbulbLight(15, 15, 10);

        // 箱
        const size = 3;
        this.addCube(0xff0000, size, -8);
        this.addCube(0x00ff00, size, 0);
        this.addCube(0x0000ff, size, 8);

        // シェーダー用テストオブジェクト
        const vertexPositions = new Float32Array([
            -10, 0, 0,
            0, 17, 0,
            10, 0, 0,
        ]);
        const vertexColors = new Float32Array([
            1, 0, 0,
            0, 1, 0,
            0, 0, 1,
        ]);
        const boardGeometry = new THREE.BufferGeometry();
        boardGeometry.setAttribute(
            "position",
            new THREE.BufferAttribute(vertexPositions, 3)
        );
        boardGeometry.setAttribute(
            "color",
            new THREE.BufferAttribute(vertexColors, 3)
        );
        const boardMaterial = new THREE.ShaderMaterial({
            vertexShader: vertShaderExample,
            fragmentShader: fragShaderExample,
            uniforms: {
                time: { value: 0 },
            },
            defines: {
                PI2THIRD: Math.PI * 2 / 3,
            },
            side: THREE.DoubleSide,
        });
        this.colorBoard = new THREE.Mesh(boardGeometry, boardMaterial);
        this.colorBoard.position.set(0, 0, -20);
        this.scene.add(this.colorBoard);
    }

    addbulbLight(x: number, y: number, z: number) {
        const bulbGeometry = new THREE.SphereBufferGeometry(0.1, 16, 8);
        const bulbLight = new THREE.PointLight(0xffee88, 300, 100, 2);
        const bulbMat = new THREE.MeshStandardMaterial({
            emissive: 0xffffee,
            color: 0x000000
        });
        bulbLight.add(new THREE.Mesh(bulbGeometry, bulbMat));
        bulbLight.position.set(x, y, z);
        bulbLight.castShadow = true;
        this.scene.add(bulbLight);
    }

    addCube(color: number, size: number, x: number) {
        const boxGeometry = new THREE.BoxGeometry(size, size, size);
        const boxMaterial = new THREE.MeshPhongMaterial({
            color: color
        });
        const cube = new THREE.Mesh(boxGeometry, boxMaterial);
        cube.castShadow = true;
        cube.position.set(x, size / 2, 0);
        cube.rotateY(Math.PI / 4);
        this.scene.add(cube);
    }

    animate() {
        this.stats.begin();
        this.controls.update();
        this.renderer.render(this.scene, this.camera);
        // シェーダーのアニメーションタイムステップを増加
        this.colorBoard.material.uniforms.time.value++;
        this.stats.end();

        requestAnimationFrame(() => this.animate());
    }
}

const test = new TestClass();
test.animate();
